;
;     Project: hello world SPI SRAM test for Effects boards
;
; Description: SPI SRAM data reliability diagnostic
;              Interrupts and DMA is used.  
;              This program is designed to test one 23K256 SRAM.
;
; SPI SRAM uses pin 15 (RB6) to operate ~CS pin.
; UART1 uses Pin 21 RP10 as output.
;
; DMA operation requires two DMA channels, one for transmit and one for
; receive.
;  channel 0  -  transmit
;  channel 1  -  receive
;
; FROM SPI PDF:
; RX and TX in Master mode: In this configuration, no DMA request is issued until the first
; block of SPI data is received. However, in Master mode, no data is received until the SPI
; transmits it. To initiate DMA transfers, the user application must first send data using the
; DMA Manual Transfer mode, or it must first write data into the SPI buffer independently of
; the DMA.
;
; Since we need to control exactly when data is transmit/received to/from SRAM, the DMA 
; transfers must be started manually.
;
;
; NOTE: DMA One-shot mode causes the CHEN bit to be reset when the DMA request is completed.
;       Because of this, it is best to set CHEN and then set FORCE each time a DMA request
;       is needed.
;
; WORKS AS OF 2013-01-14
;
;------------------------------------------------------------------------------

;------------------------------------------------------------------------------
;Sec 0.0. Processor Setup
;------------------------------------------------------------------------------
 
.equ __33F128GP802, 1
.include "p33FJ128GP802.inc"
 
;------------------------------------------------------------------------------
;Sec 0.1.Configuration Fuses
;------------------------------------------------------------------------------

.include "p33FJ128GP802_40MIPS_20MHZxtal.fuse" 

;------------------------------------------------------------------------------
;Sec 0.2.Global Equates & Symbols
;------------------------------------------------------------------------------
 
.equiv LED, 4                     ; there is a LED on PORTA, #4
 
;------------------------------------------------------------------------------
;Sec 0.3.Global Declarations
;------------------------------------------------------------------------------
.global     __reset          ; reset vector
.global     __DMA1Interrupt  ; DMA channel 1 (read from SPI)
.global     __DMA0Interrupt  ; DMA channel 0 (write to SPI)
  
;------------------------------------------------------------------------------
;Sec 0.4.Global Initialised Variables
;------------------------------------------------------------------------------
.section .data

  SPI1_MSG_COMPLETE:      .word 0x0000
 
  LFSR_HI:                .word 0x0000
  LFSR_LO:                .word 0x0000

.section *,bss,dma

; this is storage for the outgoing message stream (to SPI SRAM)
  SPI1_WRITE_BUF:
  SPI1_CMD:               .space 1
  SPI1_ADDR_HI:           .space 1
  SPI1_ADDR_LO:           .space 1
  SPI1_WRITE_DATA:        .space 1         ; store data to write here (only matters for write command

; this is storage for the incoming message stream (from SPI SRAM)
  SPI1_READ_BUF:
  SPI1_dummy0:            .space 1
  SPI1_dummy1:            .space 1
  SPI1_dummy2:            .space 1
  SPI1_READ_DATA:         .space 1         ; data read from SPI SRAM is stored here


  NOT_SURE_WHY_I_NEED_THIS: .space 64

;------------------------------------------------------------------------------
;Sec 1.0.Main Code
;------------------------------------------------------------------------------

.section .text

__reset:

.include "40MIPS.startup"

.include "SPI1_MASTER_to_SRAM_DMA.init"

.include "UART1.init"
  
  CALL LFCR
  MOV #'.', W1
  CALL TTYOUT
  MOV #'.', W1
  CALL TTYOUT
  MOV #'.', W1
  CALL TTYOUT
  CALL LFCR

; initialize internal SRAM
  
  CLR SPI1_MSG_COMPLETE
  
  
; initialize LFSR:
  MOV #0xF928, W0
  MOV W0, LFSR_HI
  MOV #0x28D1, W0
  MOV W0, LFSR_LO

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

  BCLR TRISA, #TRISA4             ; use RA4 for an output for LED

  BCLR TRISB, #6                  ; Pin 15 (RB6) is used to operate ~CS line
  BSET PORTB, #6                  ; set RB6 high (SPI SRAM inactive)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; DMA notes
; FOR READS AND WRITES:
; - start DMA READ operation first (nothing will happen until the write operation starts
;   but this ensures coordination and prevents data loss)
; - start DMA WRITE operation
;   This starts the clock pulses
;
; This does indeed look a bit weird, however, reads will not work without writes because
; the clock toggles only during write.  During a DMA READ, a read command is issued and the SRAM
; will send valid data on it's SDO line.  The writes are used to send the command and address
; and a dummy byte to clock the data out of the SRAM.  During a DMA WRITE, the write command is 
; issued and the SRAM will send junk data on it's SDO line, but we still need to read it (and then
; ignore whatever it sent) because the SPI controller gets messed up if we don't pretend we need
; the data.  Ultimately, the read and write processes are almost identical, except that when we 
; write, we supply the real byte to write and ignore what was read.  For reads, we send a dummy
; data byte and use the data that came from the SRAM.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; write 32768 bytes of random data
; reset the random generator
; read 32768 bytes of RAM and compare to original random data.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

  CLR W13                         ; W13 is used to collect the read data byte for comparison.  As long
                                  ; as the upper byte is never written, it will remain zero

  BCLR SPI1_MSG_COMPLETE, #0      ; set the message complete flag to zero
  
  BCLR PORTA, #LED

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;-- MAIN LOOP --------------------------------------------------------
diag_loop:

;  BTG PORTA, #LED                 ; if this is working, the LED should be not fully on and not fully off

  CLR W11                         ; W11 has RAM address

  MOV LFSR_HI, W8
  MOV LFSR_LO, W9                 ; save state of LFSR

;-- WRITE LOOP --------------------------------------------------------
write_loop:                       ; write 32K of random data
  CALL RAND                       ; do LFSR shift/feedback
  MOV LFSR_LO, W12                ; W12 has data to write

  CALL SRAM_WRITE                 ; THIS STARTS DMA WRITE OPERATION - data is in W12, address in W11

;  BTG PORTA, #LED                 ; if this is working, the LED should be not fully on and not fully off

  INC W11, W11                    ; inc address to next SPI SRAM byte

;-- WAIT FOR INTERRUPT  --------------------------------------------------------
; now wait for interrupt (because this is a RAM diagnostic, in applications, we could do other stuff
WAIT_FOR_WRITE_INTERRUPT:
  BTG PORTA, #LED              ; toggle the LED at each interrupt
  BTST SPI1_MSG_COMPLETE, #0      ; wait for bit 0 to go high
  BRA Z, WAIT_FOR_WRITE_INTERRUPT ; loop until it goes high
  BCLR SPI1_MSG_COMPLETE, #0      ; clear the message complete flag

  MOV #0x8000, W0                 ;
  CP W0, W11                      ; compare new address with 0x8000
  BRA NZ, write_loop              ; if it isn't equal, then we're not done yet


;----------------------------------------------------------
;----------------------------------------------------------
  MOV W8, LFSR_HI                 ; restore state of LFSR to that before starting the write loop
  MOV W9, LFSR_LO

  CLR W11                         ; set address to zero

read_loop:
  CALL RAND                       ; do LFSR shift/feedback
  MOV LFSR_LO, W12                ; use low word for random data

  CALL SRAM_READ                  ; THIS STARTS DMA READ OPERATION - read SRAM data into W13

WAIT_FOR_READ_INTERRUPT:
  BTG PORTA, #LED              ; toggle the LED at each interrupt
  BTST SPI1_MSG_COMPLETE, #0      ; wait for bit 0 to go high
  BRA Z, WAIT_FOR_READ_INTERRUPT
  BCLR SPI1_MSG_COMPLETE, #0      ;

  MOV #SPI1_READ_DATA, W3

  CP.B W12, [W3]                    ; compare what was read from SRAM to what was generated.
  BRA Z, no_error



; we are here because the data read from SRAM does not match the random data designated for this location.
  MOV W11, W3                     ; W11 has address
  CALL HEXwordOUT
  MOV #' ', W1
  CALL TTYOUT

  MOV W12, W3                     ; W12 has correct data
  CALL HEXwordOUT
  MOV #' ', W1
  CALL TTYOUT

  MOV W13, W3                     ; W13 has data read from SRAM
  CALL HEXwordOUT
  CALL LFCR

no_error:
  INC W11, W11                    ; inc address to next SPI SRAM byte

; now wait for interrupt (because this is a RAM diagnostic, in applications, we could do other stuff

  MOV #0x8000, W0                 ; if it isn't equal, then we're not done yet
  CP W0, W11                      ; compare new address with 0x8000
  BRA NZ, read_loop               ; if it isn't equal, then we're not done yet

  MOV #'+', W1                    ; mark a completed pass to serial character out
  CALL TTYOUT

  GOTO diag_loop

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Subroutines

SRAM_READ:                 ; address is in W11
  MOV #SPI1_WRITE_BUF, W13 ; pointer to write buffer
  MOV #3, W0               ; read command
  MOV.B W0, [W13++]        ;
  SWAP W11
  MOV W11, W0
  MOV.B W0, [W13++]        ;
  SWAP W11
  MOV W11, W0
  MOV.B W0, [W13++]        ;
  BCLR PORTB, #6           ; clear CS
  BSET DMA1CON, #CHEN      ; enable read channel
  BSET DMA1CON, #FORCE     ; start read request
  BSET DMA0CON, #CHEN      ; enable write channel
  BSET DMA0REQ, #FORCE     ; start write request
  RETURN

SRAM_WRITE:                ; address is in W11
  MOV #SPI1_WRITE_BUF, W13 ; pointer to write buffer
  MOV #2, W0               ; write command
  MOV.B W0, [W13++]        ;
  SWAP W11
  MOV W11, W0
  MOV.B W0, [W13++]        ;
  SWAP W11
  MOV W11, W0
  MOV.B W0, [W13++]        ;
  MOV W12, W0
  MOV.B W0, [W13++]        ; W12 has byte to write to SPI SRAM
  BCLR PORTB, #6           ; clear CS
  BSET DMA1CON, #CHEN      ; enable read channel
  BSET DMA1CON, #FORCE     ; start read request
  BSET DMA0CON, #CHEN      ; enable write channel
  BSET DMA0REQ, #FORCE     ; start write request
  RETURN

HEXwordOUT:                ; print hex value of W3
  PUSH W2
  PUSH W3
  MOV W3, W2
  SWAP W2
  CALL HEXbyteOUT
  MOV W3, W2
  CALL HEXbyteOUT
  POP W3
  POP W2
  RETURN

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

RAND:
; operate LFSR:     32,22,2,1  (31,21,1,0)  [15,5][1,0]
  PUSH W8
  MOV #LFSR_LO, W8
  MOV #1, W0             ; toggler for implementing feedback XNOR
  BTSC LFSR_HI, #15      ; test bit 15 and skip if clear
  BTG W0, #0
  BTSC LFSR_HI, #5
  BTG W0, #0
  BTSC LFSR_LO, #1
  BTG W0, #0
  BTSC LFSR_LO, #0
  BTG W0, #0
; W0 bit 0 is the XNOR output
  SL LFSR_LO             ; shift left low word with high bit to carry
  RLC LFSR_HI            ; rotate left hi word inserting carry as low bit (high bit goes to carry)
  IOR W0, [W8], [W8]     ; IOR W0 with LFSR_LO, store in LFSR_LO
; LFSR is shifted and ready.
  POP W8
  RETURN

HEXbyteOUT:              ; print hex value of W2
  PUSH W0
  PUSH W1
  PUSH W2
  MOV W2, W0
  SWAP.B W0          ; swap nybbles in lower byte
  AND #0x000F, W0
  CP W0, #10
  BRA GE, ALPH1
  IOR #0x0030, W0
  GOTO HBO1  
ALPH1:
  ADD #0x0037, W0
HBO1:
  MOV W0, W1
  CALL TTYOUT
  MOV W2, W0
  AND #0x000F, W0
  CP W0, #10
  BRA GE, ALPH2
  IOR #0x0030, W0
  GOTO HBO2
ALPH2:
  ADD #0x0037, W0
HBO2:  
  MOV W0, W1
  CALL TTYOUT
  POP W2
  POP W1
  POP W0
  RETURN  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

TTYOUT:              ; Character to send is in W1
  BTST U1STA, #8     ; test TRMT bit "transmit shift register is empty"
  BRA Z, TTYOUT      ; wait until this is high
  MOV W1, U1TXREG    ; 
  RETURN  

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

LFCR:
  PUSH W1
  MOV #0x000A, W1
  CALL TTYOUT
  MOV #0x000D, W1
  CALL TTYOUT
  POP W1
  RETURN

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

timer:
;  MOV #5, W1
  MOV #50, W1
timer1:
;  MOV #1000, W0
  MOV #10000, W0
timer2:
  DEC W0, W0
  BRA NZ, timer2

  DEC W1, W1
  BRA NZ, timer1
  RETURN
     

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; DMA0 will not be set to interrupt.  This is because for every byte written, a byte
; will be read.  When the transfer is complete, we will manage the DMA0 DMA controller
; from within the __DMA1Interrupt ISR
__DMA1Interrupt:               ; interrupt vector for DMA read transfer from SPI SRAM
  BSET PORTB, #6               ; set CS high
  BSET SPI1_MSG_COMPLETE, #0   ; set the message complete flag
  BCLR IFS0, #DMA1IF           ; clear the interrupt flag for DMA1
  RETFIE

.end
